function recorder
%RECORDER
% Recorder tuning demo

% Copyright 2010-2011 Peter Fiala
% Laboratory of Acoustics
% Dept. of Telecommunications
% Budapest University of Technology and Economics

% Build gui Figure with default parameters
MainFigure = figure('Visible', 'on',...
    'Name', 'Recorder Tuner', 'NumberTitle', 'off',...
    'Units', 'characters',...
    'Color', get(0,'defaultUicontrolBackgroundColor'), ...
    'Units', 'pixels',...
    'Position',[0 0 1000 700],...
    'MenuBar', 'none',...
    'Toolbar', 'Figure');

% GUI axes
axes('Units', 'normalized', 'Tag', 'axRec', 'NextPlot', 'ReplaceChildren', ...
    'Position', [.05 .1, .1, .8], 'YDir', 'reverse');
axes('Units', 'normalized', 'Tag', 'axNet', 'NextPlot', 'ReplaceChildren', ...
    'Position', [.2 .1, .2, .8], 'YDir', 'reverse');
axis off;
axes('Units', 'normalized', 'Tag', 'axImp', 'NextPlot', 'ReplaceChildren', ...
    'Position', [.45 .1, .5, .4]);

% GUI input panels
panelGeom = uipanel(MainFigure, 'Title', 'Recorder Geometry', ...
    'Position', [.45, .75, .16, .2]);
uidata = {
    'Length [m]', '2e-1', 'L', @(src,data)editGeom_Callback(src);
    'Inner radius [m]', '1e-2', 'R', @(src,data)editGeom_Callback(src);
    'Wall thickness [m]', '5e-3', 'w', @(src,data)editGeom_Callback(src);
    };
print_panel(panelGeom, uidata);

panelHoles = uipanel(MainFigure, 'Title', 'Sound Holes', ...
    'Position', [.62, .75, .32, .2]);
uidata = {
    'Position [m]', '[.1, .12, .13, .14, .15, .16, .17]', 'HolePos', '';
    'Radius [m]', '[3e-3, 3e-3, 3e-3, 3e-3, 3e-3, 3e-3, 3e-3]', 'HoleRad', '';
    };
print_panel(panelHoles, uidata);

% GUI buttons
uicontrol(MainFigure, 'Style', 'PushButton', 'Tag', 'butUpdate', ...
    'Units', 'Normalized', 'Position', [.55, .57, .09, .05],...
    'Callback', @(src,data)RecorderUpdate(src), 'String', 'Update Recorder');
uicontrol(MainFigure, 'Style', 'PushButton', 'Tag', 'butPlot', ...
    'Units', 'Normalized', 'Position', [.65, .57, .09, .05],...
    'Callback', @(src,data)PlotFunction(src), 'String', 'Plot Admittance');

% Initialize data structure
data.open = [];
data.nodes = true;
guidata(MainFigure, data);

% initialize GUI by pressing Update button
RecorderUpdate(MainFigure);
end

function RecorderUpdate(hObject)
handles = guihandles(hObject);
data = guidata(hObject);

% Read Material parameters from GUI
geom = ReadGeometry(hObject);
holes = ReadHoles(hObject);

% open and nodes data
nHoles = length(holes.HolePos);
nNodes = 3*nHoles+3;
if length(data.open) < nHoles
    data.open(nHoles) = false;
end
data.open = data.open(1:nHoles);
if length(data.nodes) < nNodes
    data.nodes(nNodes) = false;
end
data.nodes = data.nodes(1:nNodes);

% Compute and store pipe data
data.pipe_struct = pipe_struct(geom.L, geom.R, geom.w, holes.HolePos, holes.HoleRad);

% Plot pipe model and network
data.hHoles = plot_flute(handles.axRec, geom.L, geom.R, geom.w,...
    holes.HolePos, holes.HoleRad, @(src,data,i)HoleHit(src,i));
[data.hPipes, data.hTerms, data.hNodes] = plot_pipe_net(handles.axNet,...
    length(holes.HolePos), @(src,data,i)HoleHit(src,i), @(src,data,i)NodeHit(src,i));
guidata(hObject, data);

update_holes(data);
update_nodes(data);
end

function PlotFunction(hObject)
csound = 340;

data = guidata(hObject);
handles = guihandles(hObject);

geom = ReadGeometry(hObject);
F0 = csound/(2*geom.L);
fvec = linspace(0, 8*F0, 3001);

p0 = close_hole(data.pipe_struct, 1:length(data.open));
Zin0 = Zin_pipe(p0, 2*pi*fvec/csound);
f0 = first_max(fvec, 1./abs(Zin0(1,:)));

p = close_hole(data.pipe_struct, find(~data.open));
Zin = Zin_pipe(p, 2*pi*fvec/csound);
Hin = 20*log10(abs(1./Zin(1,:)));
[f1, v1] = first_max(fvec, Hin);

n = find(data.nodes);

if ~isempty(n)
    plot(handles.axImp, fvec, 20*log10(abs(1./(Zin(n,:)))));
    t = text(f1, v1, sprintf('%.2f cent', 1200*log2(f1/f0)), 'parent', handles.axImp);
    set(t, 'FontSize', 12, 'FontWeight', 'bold');
    legend(handles.axImp, num2str(n'));
    xlabel(handles.axImp, 'Frequency [Hz]');
    ylabel(handles.axImp, 'Admittance [dB]');
    grid(handles.axImp, 'on');
end
end

function HoleHit(hObject, i)
data = guidata(hObject);
data.open(i) = ~data.open(i);
guidata(hObject, data);

update_holes(data);
update_nodes(data);
end

function update_holes(data)
col_inactive = [.5 .5 .5];
col_active = [1 1 1];

for i = 1 : length(data.open)
    if data.open(i)
        set(data.hHoles(i), 'FaceColor', 'white');
        set(data.hPipes(i), 'FaceColor', col_active);
        set(data.hTerms(i), 'FaceColor', col_active);
    else
        set(data.hHoles(i), 'FaceColor', 'black');
        set(data.hPipes(i), 'FaceColor', col_inactive);
        set(data.hTerms(i), 'FaceColor', col_inactive);
    end
end
end

function NodeHit(hObject, i)
data = guidata(hObject);
data.nodes(i) = ~data.nodes(i);
guidata(hObject, data);

update_nodes(data);
end

function update_nodes(data)
col_inactive = 'black';
col_active = 'red';

for i = 1 : length(data.hNodes)
    if data.hNodes(i)
        if data.nodes(i)
            set(data.hNodes(i), 'MarkerFaceColor', col_active);
        else
            set(data.hNodes(i), 'MarkerFaceColor', col_inactive);
        end
    end
end
end

function geom = ReadGeometry(hObject)
handles = guihandles(hObject);
geom.L = str2double(get(handles.editL, 'String'));
geom.R = str2double(get(handles.editR, 'String'));
geom.w = str2double(get(handles.editw, 'String'));
end

function holes = ReadHoles(hObject)
handles = guihandles(hObject);
holes.HolePos = eval(get(handles.editHolePos, 'String'));
holes.HoleRad = eval(get(handles.editHoleRad, 'String'));
end

function print_panel(panel, uidata)
h = 1.5;
h0 = .7;
n = size(uidata,1);
Height = n*h+(n+2)*h0;
D = 30;
d = 20;

set(panel, 'Units', 'characters');
Pos = get(panel, 'Position');
Pos(2) = Pos(2)+Pos(4)-Height;
Pos(4) = Height;
set(panel, 'Position', Pos);

for i = 1 : size(uidata,1)
    uicontrol(panel, ...
        'Style', 'Text',...
        'Units', 'characters', ...
        'String', uidata{i,1},...
        'HorizontalAlignment', 'left',...
        'Position', [2*h0, Height-h0-i*(h+h0), D, h],...
        'Units', 'normalized');

    uicontrol(panel, ...
        'Style', 'Edit',...
        'Tag', sprintf('edit%s', uidata{i,3}),...
        'Units', 'characters', ...
        'Position', [d, Height-h0-i*(h+h0), Pos(3)-d-h, h],...
        'Callback', uidata{i,4}, ...
        'BackGroundColor', [1 1 1], ...
        'String', uidata{i,2},...
        'Units', 'normalized');
end
set(panel, 'Units', 'normalized');
end

function editGeom_Callback(hObject)
uidata = {
    'L', [0 Inf], 1e-1;
    'R', [0 Inf], 1e-2;
    'w', [0 Inf], 5e-3;
    };
checkdefault(hObject, uidata);
end

function checkdefault(hObject, data)
handles = guihandles(hObject);
for i = 1 : size(data,1)
    htag = sprintf('edit%s', data{i,1});
    lim = data{i,2};
    def = data{i,3};
    str = get(handles.(htag), 'String');
    num = str2double(str);
    if isnan(num)
        num = def;
    else
        if num < lim(1)
            num = lim(1);
        end
        if num > lim(2)
            num = lim(2);
        end
    end
    set(handles.(htag), 'String', num2str(num, '%g'));
end
end

function f = plot_flute(h, L, R, w, hole_pos, hole_rad, fhandle)
cla(h);
hold(h, 'on');
patch([-(R+w); (R+w); (R+w); -(R+w)], [0; 0; L; L], [0 0 0], 'parent', h);
patch([-(R); (R); (R); -(R)], [0; 0; L; L], [.7 .7 .7], 'parent', h);
phi = (0:15)/16*2*pi;
f = zeros(1,length(hole_pos));
for i = 1 : length(hole_pos)
    f(i) = fill(hole_rad(i)*cos(phi),...
        hole_pos(i)+hole_rad(i)*sin(phi), [0 0 0], ...
        'ButtonDownFcn', {fhandle, i}, 'parent', h);
end
hold(h, 'off');
axis(h, 'equal', 'tight');
set(h, 'YDir', 'reverse');
end

function [f, g, e] = plot_pipe_net(h, nHoles, fhandle, fhandle2)
lPipe = 1;
dPipe = .4;
sPipe = .4;
lTerm = .5;
dTerm = .5;

col_inactive = [.5 .5 .5];
col_active = [1 1 1];

cla(h);
hold(h, 'on');

line([0;0], ((nHoles+1)*(lPipe+2*sPipe)+2*sPipe+lTerm)*[0;1],...
    'Color', 'black', 'parent', h);

for i = 0 : nHoles
    patch([-dPipe/2, dPipe/2, dPipe/2, -dPipe/2], ...
        i*(2*sPipe+lPipe)+sPipe + [0 0 lPipe lPipe], col_active, 'parent', h);
end

text(0, 0, [num2str(1) '  '], 'HorizontalAlignment', 'right', 'parent', h);
e(1) = plot(h, 0, 0, 'ko', 'MarkerFaceColor', 'black', 'MarkerSize', 7, ...
    'ButtonDownFcn', {fhandle2, 1});

f = zeros(nHoles,1);
g = zeros(nHoles,1);
for i = 1 : nHoles
    line([0;4*sPipe+lPipe+lTerm], i*(lPipe+2*sPipe)*[1;1],...
        'Color', 'black', 'parent', h);
    ind = 1+3*i-2;
    text(0, i*(2*sPipe+lPipe), [num2str(ind) '  '], ...
        'HorizontalAlignment', 'right', 'parent', h);
    e(ind) = plot(h, 0, i*(2*sPipe+lPipe), 'ko',...
        'MarkerFaceColor', 'black',...
        'MarkerSize', 7, 'ButtonDownFcn', {fhandle2, ind});
    f(i) = patch(sPipe + [0 0 lPipe lPipe],...
        i*(2*sPipe+lPipe)+[-dPipe/2, dPipe/2, dPipe/2, -dPipe/2],...
        col_inactive, 'ButtonDownFcn', {fhandle, i}, 'parent', h);
    g(i) = patch(3*sPipe + lPipe + [0 0 lTerm lTerm],...
        i*(2*sPipe+lPipe)+[-dTerm/2, dTerm/2, dTerm/2, -dTerm/2],...
        col_inactive, 'ButtonDownFcn', {fhandle, i}, 'parent', h);
    ind = 1+3*i-1;
    text(2*sPipe+lPipe, i*(2*sPipe+lPipe)-0.3, num2str(ind), ...
        'HorizontalAlignment', 'center', 'parent', h);
    e(ind) = plot(h, 2*sPipe+lPipe, i*(2*sPipe+lPipe), 'ko',...
        'MarkerFaceColor', 'black', 'MarkerSize', 7,...
        'ButtonDownFcn', {fhandle2, ind});
    plot(h, 2*sPipe+lPipe+2*sPipe+lTerm, i*(2*sPipe+lPipe), 'ko',...
        'MarkerFaceColor', 'black', 'MarkerSize', 7);
end
ind = 1+3*(nHoles+1)-2;
text(0, (nHoles+1)*(2*sPipe+lPipe), [num2str(ind) '  '],...
    'HorizontalAlignment', 'right', 'parent', h);
e(ind) = plot(h, 0, (nHoles+1)*(2*sPipe+lPipe), 'ko', 'MarkerFaceColor', 'black',...
    'MarkerSize', 7, 'ButtonDownFcn', {fhandle2, ind});
patch([-dTerm/2, dTerm/2, dTerm/2, -dTerm/2],...
    (nHoles+1)*(2*sPipe+lPipe)+sPipe+[0 0 lTerm lTerm], col_active, 'parent', h);
plot(h, 0, (nHoles+1)*(2*sPipe+lPipe)+2*sPipe+lTerm, 'ko',...
    'MarkerFaceColor', 'black', 'MarkerSize', 7);

hold(h, 'off');
axis(h, 'equal',  'tight');
ylim(h, [-.1 max(ylim(h))+.1]);
end

function pipe_struct = pipe_struct(L, R, w, hole_pos, hole_rad)
nHole = length(hole_pos);
nDOF = 3*nHole+1+2;

[hole_pos, i] = sort(hole_pos);
hole_rad = hole_rad(i);

% longitudinal pipe segments
len = diff([0 hole_pos L]);
left = [1 3*(1:length(len)-1)-1];
right = [left(2:end) left(end)+3];
pipe_struct = [left' right' ones(nHole+1,1) len' repmat(R^2*pi, length(len),1)];

% hole pipes
left = left(2:end);
pipe_struct = [
    pipe_struct
    left' left'+1 ones(nHole,1) repmat(w, length(left),1), hole_rad'.^2*pi
    ];

% hole terminators
left = left + 1;
pipe_struct = [
    pipe_struct
    left' left'+1 2*ones(nHole,1) hole_rad', zeros(nHole,1)
    ];

% pipe terminator
pipe_struct = [
    pipe_struct
    nDOF-1 nDOF 2 R 0
    ];
end

function [fmax, dmax] = first_max(f, data)
d1 = diff(data(:,1:end-1), 1, 2);
d2 = diff(data(:,2:end), 1, 2);
fmax = zeros(size(data,1),1);
dmax = fmax;
for i = 1 : size(data,1)
    [trash, ind] = find(d1(i,:) > 0 & d2(i,:) < 0, 1, 'first');
    fmax(i,1) = f(ind+1);
    dmax(i,1) = data(i,ind+1);
end
end

function Zin = Zin_pipe(pipe_data, kvec)

% Allocate space for impedances
nDOF = max(max(pipe_data(:,1:2)));
nK = length(kvec);
Zin = zeros(nDOF,nK);
ready = false(nDOF,1);

% find terminating nodes and set impedance to zero
term = pipe_data(~ismember(pipe_data(:,2), pipe_data(:,1)), 2);
ready(term) = true;

%
while any(~ready)
    for left = 1 : nDOF
        if ready(left)
            continue;
        end
        e = find(pipe_data(:,1) == left);
        right = pipe_data(e,2);
        if all(ready(right))
            H = zeros(1, nK);
            for iE = 1 : length(e)
                H = H + 1./Zin_elem(pipe_data(e(iE),3:end), Zin(right(iE),:), kvec);
            end
            Zin(left,:) = 1./H;
            ready(left) = true;
        end
    end
end

end

function Zin = Zin_elem(elemdata, Zt, k)
% ELEMDATA : [1 L A] for pipe segment
%            [2 R]   for terminating element
% Zt terminating acoustic impedance
% k wave number

csound = 340;
rho0 = 1.2;
z0 = csound * rho0;

switch elemdata(1)
    case 1 % pipe
        L = elemdata(2);
        A = elemdata(3);
        Z0 = z0 / A;
        kL = k * L;
        ckL = cos(kL);
        skL = sin(kL);
        Zin = Z0 * (Zt.*ckL + 1i*Z0*skL) ./ (1i*Zt.*skL + Z0*ckL);
    case 2  % terminating impedance (half space)
        R = elemdata(2);
        A = R^2*pi;
        Z0 = z0 / A;
        Zin = Z0 * halfspaceimpedance(k*R);
%         Z = Z0 * ((kR).^2/2 - (kR).^4/(2^2*3) + (kR).^6/(2^2*3^2*4));
%         X = 4*Z0/pi * ((2*kR)/3 - (2*kR).^3/(3^2*5) + (2*kR).^5/(3^2*5^2*7));
%         Zin = Z + 1i*X;
end
end

function p = close_hole(pipe, holes)
nElem = size(pipe,1);
nHole = (nElem-2)/3;
ind1 = nHole+1 + holes;
ind2 = nHole+1 + nHole + holes;
p = pipe(setdiff(1:nElem, unique([ind1 ind2])),:);
end

function z = halfspaceimpedance(gamma)
R = 1-(besselj(1,2*gamma)./gamma);
X = zeros(size(gamma));
x = -2./(pi*gamma);
q = (2*gamma).^2;
for n = 1 : 500
    x = x .* (-q/(2*n-1)/(2*n+1));
    X = X + x;
end
z = R + 1i*X;
end